from enum import Enum

from Crypto.Cipher import AES
from home_assistant_bluetooth import BluetoothServiceInfo


class TelegramType(Enum):
    DATA = 0
    COMMISSION = 1


ALL_TELEGRAM_LENGTH_FIELD_SIZE: int = 1
ALL_TELEGRAM_TYPE_FIELD_SIZE: int = 1
ALL_TELEGRAM_MANUFACTURER_ID_SIZE: int = 2

DATA_TELEGRAM_SEQ_COUNTER_SIZE: int = 4
DATA_TELEGRAM_SWITCH_STATUS_SIZE: int = 1
DATA_TELEGRAM_SIGNATURE_SIZE: int = 4

ALL_TELEGRAM_TYPE_VALUE: bytes = b"\xFF"


class PTM215BDecoder:
    def __init__(self, raw_data: BluetoothServiceInfo) -> None:
        self.manufacturer_id = raw_data.manufacturer_id
        self.manufacturer_data = raw_data.manufacturer_data[raw_data.manufacturer_id]
        self.source_address = raw_data.address

        # Since we do not have the complete telegram at hand
        # we need to recontruct telegram total length to decide on telegram_type
        generic_field_bytes = (
            ALL_TELEGRAM_LENGTH_FIELD_SIZE
            + ALL_TELEGRAM_TYPE_FIELD_SIZE
            + ALL_TELEGRAM_MANUFACTURER_ID_SIZE
        )
        telegram_bytes_total = len(self.manufacturer_data) + generic_field_bytes

        if telegram_bytes_total == 30:
            self.telegram_type = TelegramType.COMMISSION
        else:
            self.telegram_type = TelegramType.DATA

    def is_signature_valid(self, sec_key: bytes) -> bool:
        if self.telegram_type == TelegramType.COMMISSION:
            raise ValueError("Signature check only possible for Data Telegrams.")

        # Calculate signature and verify against submitted signature
        cipher = AES.new(
            sec_key,
            AES.MODE_CCM,
            nonce=self._get_nonce(),
            mac_len=4,
            msg_len=0,
            assoc_len=len(self._input_data),
        )

        cipher.update(self._input_data)
        try:
            cipher.verify(self.signature)
        except ValueError:
            return False
        return True

    def _get_nonce(self) -> bytes:
        """Get Nonce of length 13 bytes consisting of source_address and seq counter and padding"""

        return self._source_address_le + self._seq_counter_le + b"\x00" * 3

    @property
    def _manufacturer_id_le(self) -> bytes:
        return self.manufacturer_id.to_bytes(
            ALL_TELEGRAM_MANUFACTURER_ID_SIZE, "little"
        )

    @property
    def _seq_counter_le(self) -> bytes:
        return self.manufacturer_data[:DATA_TELEGRAM_SEQ_COUNTER_SIZE]

    @property
    def _source_address_le(self) -> bytes:
        return bytes.fromhex("".join(self.source_address.split(":")[::-1]))

    @property
    def _input_data(self) -> bytes:
        """Rebuild complete data telegram payload using rebuilded length field, fixed value from type field and manufacturer data wo signature"""
        manufacturer_data_wo_signature = self.manufacturer_data[
            :-DATA_TELEGRAM_SIGNATURE_SIZE
        ]
        return (
            self._length_field
            + ALL_TELEGRAM_TYPE_VALUE
            + self._manufacturer_id_le
            + manufacturer_data_wo_signature
        )

    @property
    def _length_field(self) -> bytes:
        return (
            ALL_TELEGRAM_TYPE_FIELD_SIZE
            + ALL_TELEGRAM_MANUFACTURER_ID_SIZE
            + len(self.manufacturer_data)
        ).to_bytes(ALL_TELEGRAM_LENGTH_FIELD_SIZE, "big")

    @property
    def optional_data(self) -> bytes:
        # First sum up bytes count of mandatory fields
        non_optional_data_bytes = (
            DATA_TELEGRAM_SEQ_COUNTER_SIZE
            + DATA_TELEGRAM_SWITCH_STATUS_SIZE
            + DATA_TELEGRAM_SIGNATURE_SIZE
        )

        # If there are any other fields than the mandatory fields we expect them to be optional data
        optional_data_bytes = len(self.manufacturer_data) - non_optional_data_bytes
        if optional_data_bytes > 0:
            # In case we have optional data bytes we calculate starting index and extract data from list
            optional_data_start_index = (
                DATA_TELEGRAM_SEQ_COUNTER_SIZE + DATA_TELEGRAM_SWITCH_STATUS_SIZE
            ) - 1

            optional_data = self.manufacturer_data[
                optional_data_start_index : optional_data_start_index
                + optional_data_bytes
            ]

            # If there is only one byte we make sure it is not interpreted as int by python
            if isinstance(optional_data, int):
                optional_data = optional_data.to_bytes(optional_data_bytes, "big")
            return optional_data
        return b""

    @property
    def switch_status(self) -> bytes:
        status = self.manufacturer_data[DATA_TELEGRAM_SEQ_COUNTER_SIZE]
        if isinstance(status, int):
            return status.to_bytes(1, "big")
        return status

    @property
    def is_press_action(self) -> bool:
        return (self.switch_status[0] & 1) == 1

    @property
    def a0_action(self) -> bool:
        return ((self.switch_status[0] >> 1) & 1) == 1

    @property
    def a1_action(self) -> bool:
        return ((self.switch_status[0] >> 2) & 1) == 1

    @property
    def b0_action(self) -> bool:
        return ((self.switch_status[0] >> 3) & 1) == 1

    @property
    def b1_action(self) -> bool:
        return ((self.switch_status[0] >> 4) & 1) == 1

    @property
    def signature(self) -> bytes:
        signature_start_index = (
            len(self.manufacturer_data) - DATA_TELEGRAM_SIGNATURE_SIZE
        )
        return self.manufacturer_data[signature_start_index:]



import pytest
from home_assistant_bluetooth import BluetoothServiceInfo

from enocean_ble.decoder import PTM215BDecoder


@pytest.fixture
def manufacturer_id():
    return 0x03DA


@pytest.fixture()
def manufacturer_data(manufacturer_id) -> dict:
    return {manufacturer_id: bytes.fromhex("5D04000011B2FA88FF")}


@pytest.fixture()
def manufacturer_data_w_optional_data(manufacturer_id) -> dict:
    return {manufacturer_id: bytes.fromhex("6304000011123452E05116")}


@pytest.fixture()
def decoder(request):
    raw_data = BluetoothServiceInfo(
        name="ptm216b_test",
        address="E2:15:00:00:19:B8",
        rssi=40,
        manufacturer_data=request.getfixturevalue(request.param),
        service_data={},
        service_uuids=[],
        source="",
    )
    return PTM215BDecoder(raw_data)


@pytest.mark.parametrize(
    "decoder, expected_field",
    [
        ("manufacturer_data", b"\x0c"),
        ("manufacturer_data_w_optional_data", b"\x0e"),
    ],
    indirect=["decoder"],
)
def test_length_field(decoder, expected_field):
    resulting_field = decoder._length_field
    assert resulting_field == expected_field


@pytest.mark.parametrize(
    "decoder, expected_field",
    [
        ("manufacturer_data", bytes.fromhex("0CFFDA035D04000011")),
        (
            "manufacturer_data_w_optional_data",
            bytes.fromhex("0EFFDA0363040000111234"),
        ),
    ],
    indirect=["decoder"],
)
def test_input_data(decoder, expected_field):
    resulting_field = decoder._input_data
    assert resulting_field == expected_field


@pytest.mark.parametrize(
    "decoder, expected_field",
    [
        ("manufacturer_data", b""),
        ("manufacturer_data_w_optional_data", bytes.fromhex("1112")),
    ],
    indirect=["decoder"],
)
def test_optional_data(decoder, expected_field):
    resulting_field = decoder.optional_data
    assert resulting_field == expected_field


@pytest.mark.parametrize(
    "decoder, expected_field",
    [
        ("manufacturer_data", bytes.fromhex("B819000015E2")),
        ("manufacturer_data_w_optional_data", bytes.fromhex("B819000015E2")),
    ],
    indirect=["decoder"],
)
def test_source_address_le(decoder, expected_field):
    resulting_field = decoder._source_address_le
    assert resulting_field == expected_field


@pytest.mark.parametrize(
    "decoder, expected_field",
    [
        ("manufacturer_data", bytes.fromhex("5D040000")),
        ("manufacturer_data_w_optional_data", bytes.fromhex("63040000")),
    ],
    indirect=["decoder"],
)
def test_seq_counter_le(decoder, expected_field):
    resulting_field = decoder._seq_counter_le
    assert resulting_field == expected_field


@pytest.mark.parametrize(
    "decoder, expected_field",
    [
        ("manufacturer_data", bytes.fromhex("DA03")),
        ("manufacturer_data_w_optional_data", bytes.fromhex("DA03")),
    ],
    indirect=["decoder"],
)
def test_manufacturer_id_le(decoder, expected_field):
    resulting_field = decoder._manufacturer_id_le
    assert resulting_field == expected_field


@pytest.mark.parametrize(
    "decoder, expected_field",
    [
        ("manufacturer_data", bytes.fromhex("B819000015E25D040000000000")),
        (
            "manufacturer_data_w_optional_data",
            bytes.fromhex("B819000015E263040000000000"),
        ),
    ],
    indirect=["decoder"],
)
def test_get_nonce(decoder, expected_field):
    resulting_field = decoder._get_nonce()
    assert resulting_field == expected_field



import logging
from typing import Optional

from bluetooth_data_tools import short_address
from bluetooth_sensor_state_data import BluetoothData
from home_assistant_bluetooth import BluetoothServiceInfo

from .decoder import PTM215BDecoder, TelegramType

logger = logging.getLogger(__name__)

MANUFACTURER_ID = 0x3DA


class EnoceanBluetoothDeviceData(BluetoothData):
    """Data for Enocean BLE Switch"""

    def __init__(
        self, security_key: Optional[str] = None, validate_signature: bool = False
    ) -> None:
        super().__init__()
        self._security_key = (
            bytes.fromhex(security_key) if security_key is not None else None
        )
        self._validate_signature = validate_signature

    def _start_update(self, data: BluetoothServiceInfo) -> None:
        if MANUFACTURER_ID not in data.manufacturer_data:
            logger.debug(f"Could not find manufacturer id {MANUFACTURER_ID} in data")
            return None

        decoder = PTM215BDecoder(data)
        if decoder.telegram_type != TelegramType.DATA:
            logger.debug("Only Update sensor for data telegrams")
            return None

        if self._validate_signature:
            if not decoder.is_signature_valid(self._security_key):
                logger.debug("Signature not valid.")
                return None
            logger.debug("Signature validated.")

        identifier = short_address(data.address)

        self.set_device_type("Enocean Switch")
        self.set_device_manufacturer(data.manufacturer)
        self.set_device_name(f"Enocean PTM215b {identifier}")

        self.update_binary_sensor(
            f"a0_pressed",
            decoder.a0_action and decoder.is_press_action,
            None,
            "Channel A0",
        )
        self.update_binary_sensor(
            "a1_pressed",
            decoder.a1_action and decoder.is_press_action,
            None,
            "Channel A1",
        )
        self.update_binary_sensor(
            "b0_pressed",
            decoder.b0_action and decoder.is_press_action,
            None,
            "Channel B0",
        )
        self.update_binary_sensor(
            "b1_pressed",
            decoder.b1_action and decoder.is_press_action,
            None,
            "Channel B1",
        )

    def commission(self, data: BluetoothServiceInfo) -> str:
        # TODO: Build logic to retrieve security key from device in commission mode
        ...



import asyncio
import logging

from bleak import BleakScanner
from bleak.backends.device import BLEDevice
from bleak.backends.scanner import AdvertisementData
from .parser import BluetoothServiceInfo, PTM215BDecoder

logger = logging.getLogger(__name__)


def simple_callback(device: BLEDevice, advertisement_data: AdvertisementData):
    if device.address.startswith("E2:15"):
        decoder = PTM215BDecoder(
            BluetoothServiceInfo.from_advertisement(device, advertisement_data, "")
        )
        print(device.address)
        print(
            "signature_valid: "
            + str(decoder.is_signature_valid(b'D\x18\xe4\x0c{{\x05na\xaa"\\C.\xa8\xad'))
        )
        print(f"a0: {decoder.a0_action}")
        print(f"b0: {decoder.b0_action}")
        print(f"a1: {decoder.a1_action}")
        print(f"b1: {decoder.b1_action}")
        print(f"is_press: {decoder.is_press_action}")


async def main():
    scanner = BleakScanner(simple_callback)

    while True:
        # print("(re)starting scanner")
        await scanner.start()
        await asyncio.sleep(2.0)
        await scanner.stop()


if __name__ == "__main__":
    asyncio.run(main())
